
import torch
import torchaudio
from denoiser import pretrained
from denoiser.dsp import convert_audio

from pafst.datasets import Dataset
from pafst.utils import write_json

from df.enhance import enhance, init_df, load_audio, save_audio
from df.utils import download_file
import soundfile as sf

def dfn3_denoise(in_audio_path, out_audio_path, model, df_state):
    audio, _ = load_audio(in_audio_path, sr=df_state.sr())
    enhanced = enhance(model, df_state, audio)
    
    if enhanced.shape[0] == 2:
        waveform = enhanced.mean(dim=0, keepdim=True)
    else:
        waveform = enhanced

    resampler=torchaudio.transforms.Resample(orig_freq=df_state.sr(), new_freq=16000)
    resampled = resampler(waveform).cpu().numpy()
    filepath = str(out_audio_path.absolute()).split(".")[0]+".wav"
    save_audio(filepath, resampled, 16000)
    return filepath

def fb_denoise(in_audio_path, out_audio_path, model):
    wav, sr = torchaudio.load(in_audio_path)
    wav = convert_audio(wav.cuda(), sr, model.sample_rate, model.chin)
    with torch.no_grad():
        denoised = model(wav[None])[0]
    filepath = str(out_audio_path.absolute()).split(".")[0]+".wav"
    sf.write(filepath, denoised.data.cpu().numpy().squeeze(), model.sample_rate)
    return filepath

def load_model(processor):
    if processor=="dfn":
        model, df_state, _ = init_df()
        return model, df_state, dfn3_denoise
    elif processor=="den":
        model = pretrained.dns64().cuda()
        return model, None, fb_denoise
    else:
        # g_c_n
        pass
    
def denoiser(dataset: Dataset, processor="dfn"): # den for facebook
    audios = dataset.audios
    model, df_state, proc_fun = load_model(processor)
    data = []

    for audio in audios:

        in_audio_path=str(audio.absolute())
        file_name = audio.with_name(f'{audio.stem}_{processor}{audio.suffix}')
        file_name = file_name.name

        out_audio_path = (dataset.output_path / file_name).resolve()
    
        if processor == "dfn":
            audio_path=proc_fun(in_audio_path, out_audio_path, model, df_state)
        elif processor == "den":
            audio_path=proc_fun(in_audio_path, out_audio_path, model)
        
        data.append({
                "denoised_audio_path": audio_path,
                "audio_filepath": in_audio_path,
            })

    write_json((dataset.output_path/"denoiser.json").resolve(), data)

    return data



import os
import uuid
from datetime import datetime
from pathlib import Path
from collections import defaultdict

from pafst.datasets import Dataset
from pafst.utils import write_json

from silero_vad import load_silero_vad, read_audio, get_speech_timestamps
from pydub import AudioSegment
from pyannote.audio import Pipeline
import torch


def generate_unique_filename(format="wav"):
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    unique_id = uuid.uuid4().hex
    return f"temp_audio_{timestamp}_{unique_id}.{format}"


class AudioLocator(object):
    """Represents a "frame" of audio data."""
    def __init__(self, path:str, start_time: float, duration: float):
        self.filepath: str = path
        self.start_time: float = start_time
        self.duration: float = duration



def diarization(
        dataset: Dataset,
        hf_token=None,

):
    """
    This function performs diarization (speaker separation) on the audio files,
    splits them by sentence, and then separates the audio by each speaker.

    Args:
        dataset (Dataset): Dataset instance.
        hf_token (str): Huggingface access token.

    Return:
        new_audios (list): List of new audio path.

    """
    if not hf_token:
        hf_token=os.getenv("HF_TOKEN")

    pipeline = Pipeline.from_pretrained(
        "pyannote/speaker-diarization-3.1",
        use_auth_token=hf_token)

    # check using gpu or not
    if torch.cuda.is_available():
        device = torch.device("cuda")
        print("Using GPU...")
    else:
        device = torch.device("cpu")
        print("Using CPU...")

    pipeline.to(device)
    audio_locator=[]
    seg = None
    padding = AudioSegment.silent(duration=1000)
    for audio in dataset.audios:

        seg_ = AudioSegment.from_file(str(audio))

        audio_locator.append(
                AudioLocator(
                    path=str(audio),
                    start_time=seg.duration_seconds if seg else 0.00,
                    duration=seg_.duration_seconds
                )
        )
        
        if not seg:
            seg = seg_
        else:
            seg += seg_
        seg += padding

        

    temp_file_path = dataset.output_path / generate_unique_filename()

    # create audio temp file
    seg.export(temp_file_path, format='wav')

    # diarization
    diarization_audio = pipeline(f"{temp_file_path}")

    speaker_num_list = defaultdict(int)
    audio_data=[]
    loc=0
    for i, (turn, _, speaker) in enumerate(diarization_audio.itertracks(yield_label=True)):

        locator=audio_locator[loc]
        audio_filepath=locator.filepath
        
        start_ms = turn.start * 1000
        end_ms = turn.end * 1000
        
        if not (turn.start >= locator.start_time and turn.end <= locator.start_time + locator.duration):
            loc=loc+1
            


        speaker_folder = dataset.output_path / f"speaker_{speaker}"
        speaker_folder.mkdir(parents=True, exist_ok=True)
        segment = seg[start_ms:end_ms]

        output_file_path = speaker_folder / f"{speaker}_{speaker_num_list[speaker]}.wav"
        segment.export(output_file_path, format="wav")

        speaker_num_list[speaker] += 1  # +1

        audio_data.append({
                "speaker_path":str(output_file_path),
                "audio_filepath": os.path.abspath(str(audio_filepath)),
                "start_time": round(turn.start-locator.start_time, 2),
                "end_time": round(turn.end-locator.start_time, 2)
            })
    # delete temp file
    if Path(temp_file_path).exists():
        temp_file_path.unlink()

    write_json(dataset.output_path / "diarization.json", audio_data)

    return audio_data


def vad(
        dataset: Dataset,
        min_silence_duration_ms=500,
        padding_duration_ms=200,

):
    model = load_silero_vad()
    audios = dataset.audios
    stamp = []
    audio_data = []
    j=0
    for audio in audios:
        wav = read_audio(str(audio))

        speech_timestamps = get_speech_timestamps(wav, model, min_silence_duration_ms=min_silence_duration_ms)

        for speech_timestamp in speech_timestamps:
            start = speech_timestamp['start']
            end = speech_timestamp['end']
            start /= 16000
            end /= 16000
            stamp.append((start, end))

        audio_segment = AudioSegment.from_file(audio)

        for i, (start, end) in enumerate(stamp):
            start_ms = start * 1000
            end_ms = end * 1000

            segment = audio_segment[start_ms:end_ms]
            padding = AudioSegment.silent(duration=padding_duration_ms)
            segment = padding + segment + padding

            # file_name = audio.with_name(f'{audio.stem}_{i}{audio.suffix}')
            file_name = 'segment-%003d-%003d%s' % (j,i,audio.suffix,)

            path = (dataset.output_path / file_name).resolve()
            segment.export(path, format=audio.suffix[1:])

            audio_data.append({
                "segment_path":str(path),
                "audio_filepath": os.path.abspath(str(audio)),
                "start_time": start,
                "end_time": end
            })
        j=j+1


    write_json(dataset.output_path / "vad.json", audio_data)
    return audio_data




import json
import uuid
from pathlib import Path
from datetime import datetime

def write_json(path, data):
    with open(path, "w", encoding="UTF-8") as f:
        json.dump(data, f)

def read_json(path):
    data = None
    with open(path, "r") as f:
        data = json.load(f)
    return data

def mod_json(filepath, new_parent_dir):
    new_parent_dir=str(new_parent_dir)
    data=read_json(filepath)
    for d in data:
        new_spk_path=new_parent_dir + "/" + "/".join(d["speaker_path"].split("/")[-2:])
        new_sep_path=new_parent_dir + "/" + d["audio_filepath"].split("/")[-1]
        d["speaker_path"]=new_spk_path
        d["audio_filepath"]=new_sep_path
    write_json(filepath, data)
    return
    

def mk_temp_dir():
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    unique_id = uuid.uuid4().hex
    temp_dir = Path.cwd() / f"temp_dir_{timestamp}_{unique_id}"
    temp_dir.mkdir(exist_ok=True)
    return temp_dir



from pathlib import Path

from tqdm import tqdm

from pafst.utils import write_json

from pafst.datasets import Dataset
from audio_separator.separator import Separator


def separator(
        dataset: Dataset
):
    separator = Separator(output_dir=dataset.output_path)
    separator.load_model()

    audios = dataset.audios
    audio_data = []

    bar = tqdm(audios,
               total=len(audios),
               leave=True,
               )

    for audio in bar:
        output_files = separator.separate(audio)
        target_path = Path(dataset.output_path) / Path(audio.name)

        # inst unlink
        (Path(dataset.output_path) / Path(output_files[0])).unlink()

        # save vocal
        if target_path.exists():
            target_path.unlink()
        Path(Path(dataset.output_path) / Path(output_files[1])).rename(target_path)

        audio_data.append({
                "separator_path":str(target_path),
                "audio_filepath": str(audio)
            })

    write_json(dataset.output_path/"separator.json", audio_data)

    return audio_data




import os
import json
from pathlib import Path
from tqdm import tqdm

from collections import defaultdict

import librosa
import whisper
from faster_whisper import WhisperModel
from whisper.tokenizer import LANGUAGES

from pydub import AudioSegment

from pafst.utils import write_json
from pafst.datasets import Dataset


whisper_model = {key: None for key in whisper._MODELS}

def whisper_stt(
        audio: Path,
        model_size='base',
        language=None,
):
    """
        Use the Whipser[https://github.com/openai/whisper] STT model to extract text.
        If there is no gpu or low performance, use the base model.

        Args:
            audio (Data): Audio data.
            model_size (str): Size of the whisper model.
            language (str): Language of the audio file to run STT.

        Return:
            str: Text in an audio file.

        """
    global whisper_model
    if model_size not in whisper_model.keys():
        raise ValueError(
            "[!] Invalid model size selected. Please choose one of the following sizes: tiny, base, small, medium, large.")

    if not whisper_model[model_size]:
        whisper_model[model_size] = whisper.load_model(model_size)

    result = whisper_model[model_size].transcribe(str(audio))

    return result

def process_segments(raw_segments):
    segments = []
    text=""
    for segment_chunk in raw_segments:
        chunk={}
        chunk["start"]=segment_chunk.start
        chunk["end"]=segment_chunk.end
        chunk["text"]=segment_chunk.text
        segments.append(chunk)
        text+=chunk["text"]+" "

    return segments, text

def segment_audio(audio_file, segments, output_path):
    """
    segments audio given a list of dict containing start time, end time
    """
    audio_data = []
    audio = AudioSegment.from_file(str(audio_file))
    
    for i, chunk in enumerate(segments):
        path = 'segment_%002d.wav' % (str(audio_file),i,)
        path = (Path(output_path) / Path(path)).resolve()

        start_time_ms = chunk["start"] * 1000  #  in milliseconds
        end_time_ms = chunk["end"] * 1000  # in milliseconds
        segment = audio[start_time_ms:end_time_ms]

        segment.export(path, format="wav") 

        audio_data.append({
            "segment_path":str(path), 
            "audio_filepath": os.path.abspath(str(audio_file)),
            "start_time": chunk["start"], 
            "end_time": chunk["end"] ,
            "text": chunk["text"]
        })

    return audio_data


def STT(
        dataset: Dataset,
        model_size='tiny',
        vad=True,
        language=None,
        compute_type="int8",
        faster=False
        
):
    """
    Read the audio files in the dataset, and use the stt function to extract text.
    Save the extracted text in the form of a json file.

    Args:
        dataset (Dataset): Audio dataset Class
        output_format (str): Output format, Defaults is json (json or txt)
        model_size (str): Size of the whisper model.
        language (str): Language of the audio file to run STT.
        vad (bool) = True: Performs vad, saves the audio files on timestamps detected by vad.
        compute_type (str) = "int8": Loads the model in "int8" quantization
        faster (bool) = False: if True uses faster whisper
    Returns:
        List[Dict]: List of the dictionary containing audio files and processed files, language detected by whisper and timestamps if vad set to True.

    """

    if language and language not in LANGUAGES:
            raise ValueError(
            f"[!] This language is not supported. Please select one of the language codes below\n{LANGUAGES}")

    audios = dataset.audios
    stt_dict = defaultdict(dict)

    output_path = dataset.output_path

    if faster:
        model=WhisperModel(model_size, compute_type=compute_type)
        if language:
            options = dict(language=language, beam_size=5, best_of=5)
        else:
            options = dict(beam_size=5, best_of=5)
        transcribe_options=dict(task="transcribe", **options)
    
    audio_data=[]
    audio_main=[]
    bar = tqdm(audios,
               total=len(audios),
               leave=True,
               )
    for audio in bar:
        if faster:
            raw_segments, info=model.transcribe(str(audio), **transcribe_options)
            segments, text=process_segments(raw_segments)
            language=info.language
        else:
            result=whisper_stt(audio, model_size)
            segments, text, language = result["segments"], result['text'], result["language"]
        audio_main.append({
            "asr_text": text,
            "audio_filepath": str(audio),
            "language": language
        })

        if vad:
            a_data = segment_audio(audio, segments, output_path)
            audio_data.extend(a_data)

    write_json(output_path/"asr.json", audio_main)
    if vad:
        write_json(output_path/"whisper_vad.json", audio_data)
        return audio_data
    return audio_main
    




import shutil
from pathlib import Path
from typing import Union, Dict, List

from pafst.utils import mk_temp_dir, mod_json
from pafst.datasets import Dataset
from pafst.denoisers import denoiser
from pafst.vad import vad
from pafst.diarization import diarization
from pafst.separator import separator
from pafst.stt import STT

class PAFST:
    """
        Make audio files into a dataset for TTS/STT.

        Args:
        path (str): Directory path with audio files.
        dataset_name (str, optional): Dataset name. Defaults to dataset_path's directory name.
        language (str, optional): Language using BCP 47 language tag. Defaults to 'en-us' (English)
        output_path (str): Output Directory. Defaults to './pafst_output'

        Example with quick start:
        p = PAFTS(
            path = 'your_audio_directory_path',
            output_path = 'output_path',
            hf_token="HUGGINGFACE_ACCESS_TOKEN_GOES_HERE"
        )
        p.run()

        """

    def __init__(
            self,
            path: str = None,
            dataset_name: str = None,
            language: str = None,
            output_path: str = 'pafst_output',
            hf_token: str = None
    ):

        self._hf_token=hf_token
        self.language=language
       
        self._dataset = Dataset(
            path=path,
            dataset_name=dataset_name,
            language=language,
            output_path=output_path
        )
        
    def separator(self):
        """
        input: 
            None, uses roformer to separate audio source from music source
        output: 
            List[Dict]]: This list contains the origianl file path and the processed audio path
        """
        audio_data=separator(self._dataset)
        return audio_data
    
    def vad(self, detector="webrtc", params: Dict={}):
        """
        input: 
            detector: "webrtc" (Optional) # uses "webrtcvad"
            params: {
                mode: 0-3 # aggressiveness, 0-mild, 3-very aggressive
                frame_duration: 10 or 20 or 30 # accepted duration in ms
                padding_duration: in milliseconds
            } (Optional)
        output: 
            List[Dict]]: This list contains the origianl file path and the processed audio path with vad timestamps
        """
        audio_data = []
        if detector == "webrtc":
            if params:
                audio_data = vad(self._dataset, **params)
            else:
                audio_data = vad(self._dataset)
        elif detector == "silero_vad":
            pass

        return audio_data

    def denoiser(self, processor="dfn"):
        """
        input: 
            processor: "dfn" (Optional)
                "dfn" is DeepFilterNet3, use "den" for facebook denoiser
        output: 
            List[Dict]]: This list contains the origianl file path and the processed audio path
        """
        audio_data = denoiser(self._dataset, processor=processor)
        return audio_data

    def diarization(self, hf_token=None):
        """
        input: 
            hf_token (Optional)
        output: 
            List[Dict]]: This list contains the origianl file path and the processed audio path
        """
        if hf_token:
            self._hf_token = hf_token

        audio_data = diarization(self._dataset, self._hf_token)
        return audio_data

    def stt(self, model_size='large-v3', vad=True, language="en", compute_type="float32"):
        """
        Uses faster_whisper.
        input:
            model_size = "large-v3" # 
            vad=True # performs Voice Activity Detection
            language="en"
            compute_type="float32" # the quantization size. float32 is no quantization at all
        output:
            List[Dict]]: This list contains the origianl file path and the processed audio path
        """
        if self.language:
            language=self.language

        audio_data = STT(self._dataset, model_size=model_size, 
                         vad=vad, language=language, compute_type=compute_type)
        return audio_data

    
    def _stage_process(self, process_function, *args, **kwargs):
        # Create unique temp directory
        process_function(self._dataset, *args, **kwargs)
        return

    def run(self):
        """
        Perform separation, diarization and stt in one go.
        """

        # if not processor_functions:
        #     processor_functions = ["separator", "diarization", "stt"]
        # functions = {
        #     "separator": separator,
        #     "vad": vad,
        #     "denoiser": denoiser,
        #     "diarization": diarization,
        #     "stt": STT
        # }
        
        original_output = self._dataset.output_path
        original_path = self._dataset.path
        
        # Stage 1: separator
        temp_dir1 = mk_temp_dir()
        self._dataset.output_path = temp_dir1
        self._stage_process(separator)
        self._dataset.path = temp_dir1

        # Stage 2: diarization
        temp_dir2 = mk_temp_dir()
        self._dataset.output_path = temp_dir2
        self._stage_process(diarization, hf_token=self._hf_token)

        original_output.mkdir(exist_ok=True)                
        
        for temp_dir in [temp_dir2]:
            for file in temp_dir.rglob('*'):
                if file.is_file():
                    relative_path = file.relative_to(temp_dir)
                    destination = original_output / relative_path
                    destination.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(file, destination)
        
        # Stage 3: STT

        self._dataset.path = original_output
        self._dataset.output_path=original_output
        self._stage_process(STT, vad=False)

        for file in temp_dir1.rglob('*'):
            if file.suffix[1:] in ['wav','mp3','ogg','flac']:
                relative_path = file.relative_to(temp_dir1)
                destination = original_output / relative_path
                destination.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(file, destination)
        
        dia_json_path = original_output / "diarization.json"
        mod_json(dia_json_path, original_output)

        
        
        # Cleanup
        for temp_dir in [temp_dir1, temp_dir2]:
            shutil.rmtree(temp_dir, ignore_errors=True)



from pathlib import Path

from pydub import AudioSegment

"""Supported Audio formats"""
AUDIO_FORMATS = [
    'wav',
    'mp3',
    'ogg',
    'flac'
]


def is_audio(path):
    file = Path(path)
    if file.suffix[1:] in AUDIO_FORMATS:
        return True
    return False


def get_duration(path):
    audio = AudioSegment.from_file(path)
    return audio.duration_seconds


class Dataset:
    """
    Audio Dataset Class.
    Process all audio data class.

    Args:
        path (str): Directory path with audio files.
        dataset_name (str, optional): Dataset name. Defaults to dataset_path's directory name.
        language (str, optional): Language using BCP 47 language tag. Defaults to 'en-us' (English)
        output_path (str): Output Directory. Defaults to './pafst_output'

    """

    def __init__(
            self,
            path: str = None,
            dataset_name: str = None,
            language: str = None,
            output_path: str = 'pafst_output'
    ):

        self._path = Path(path).resolve()
        self._output_path = Path(output_path)
        self._dataset_name = dataset_name
        self._language = language
        self._audios = []

        if not self._path.exists():
            raise FileNotFoundError("[!] Path does not exist")

        if not self._path.is_dir():
            raise NotADirectoryError("[!] Path is not directory.")

        if not self._dataset_name:
            self._dataset_name = self._path.name

        if not self._output_path.exists():
            self._output_path.mkdir(parents=True)

        # find audio file in path
        # self._audios = [Data(str(p)) for p in self._path.glob("**/*") if is_audio(p)]
        self._audios = [Path(p) for p in self._path.glob("**/*") if is_audio(p)]

    def __len__(self):
        return len(self._audios)

    def __getitem__(self, index):
        return self._audios[index]

    @property
    def audios(self):
        return self._audios

    @audios.setter
    def audios(self, audios):
        if isinstance(audios, list) and all(isinstance(audio, Path) for audio in audios):
            self._audios = audios
        else:
            raise ValueError("[!] The input value is not a list or the element in the list is not a Path type.")

    @property
    def dataset_name(self):
        return self._dataset_name

    @property
    def output_path(self):
        return self._output_path

    @property
    def path(self):
        return self._path

    @output_path.setter
    def output_path(self, output_path):
        if (isinstance(output_path, str) or isinstance(output_path, Path)) and Path(output_path).exists():
            self._output_path = output_path
        else:
            raise ValueError("[!] The input is wrong.")

    @path.setter
    def path(self, path):
        if (isinstance(path, str) or isinstance(path, Path)) and Path(path).exists():
            self._path = path
            self._audios = [Path(p) for p in self._path.glob("**/*") if is_audio(p)]
        else:
            raise ValueError("[!] The input is wrong.")

    def print_info(self):
        print(f'| > Dataset name : {self._dataset_name}')
        print(f'| > Path : {self._path}')
        print(f'| > language : {self._language}')
        print(f'| > Number of files : {len(self._audios)}')
        print(f'| > Total duration : {self.get_total_duration()}')

    def get_total_duration(self):
        total_duration = 0

        for audio in self._audios:
            total_duration += get_duration(audio)
        return total_duration



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

with open("requirements.txt", 'r', encoding='utf-8') as f:
    requirements = f.read()

setup(
    name='pafst',
    version='1.0.0',
    author='ashtavakra',
    author_email="vidyaaltar@gmail.com",
    description='Library That Preprocessing Audio For TTS/STT.',
    install_requires=requirements,
    license='MIT',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/prassr/PAFST',
    python_requires=">=3.10, <3.11",
    packages=find_packages(),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.10',
    ],
    keywords='speechrecognition asr voiceactivitydetection vad webrtc pafst audio denoising speaker diarization',
    py_modules=['pafst']

)
